/*
	ted: tiny [text] editor
	Copyright (C) 2009 Edward Cree <techsupport.alphasoft@googlemail.com>

	This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>

#define VERSION_MAJ	0
#define VERSION_MIN	2
#define VERSION_REV	1

typedef struct
{
	int nlines;
	char ** buf;
	char * filename;
	bool saved;
	int curline;
	int err; // cmd_ functions return errors in here
}
buffer;

#define VERSION_MSG "ted %hhu.%hhu.%hhu\n\
 Copyright (C) 2009 Edward Cree.\n\
 License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n\
 This is free software: you are free to change and redistribute it.\n\
 There is NO WARRANTY, to the extent permitted by law.\n",VERSION_MAJ, VERSION_MIN, VERSION_REV

#define GPL_MSG "ted Copyright (C) 2009 Edward Cree.\n\
 This program comes with ABSOLUTELY NO WARRANTY; for details type `help gpl-w'.\n\
 This is free software, and you are welcome to redistribute it\n\
 under certain conditions; type `help gpl-c' for details.\n"

#define USAGE_MSG "Usage:\n\
ted [<filename>] [{-s|--script}=<scriptfile>]\n\
ted {--help|--version}\n"

#define max(a,b)	((a)>(b)?(a):(b))

// function prototypes

char * getl(char *);
char * slurp(FILE *);
buffer cmd_new(char *);
buffer cmd_load(char *, bool);
buffer cmd_save(char *, buffer);
int cmd_show(char *, buffer);
int cmd_showpg(buffer);
buffer cmd_move(char *, buffer);
buffer cmd_insert(char *, buffer);
buffer cmd_replace(char *, buffer);
buffer cmd_append(char *, buffer);
buffer cmd_split(char *, buffer);
buffer cmd_join(char *, buffer);
buffer cmd_delete(char *, buffer);
int cmd_copy(char *, buffer);
buffer cmd_paste(char *, buffer);
int cmd_count(char *, buffer);
buffer cmd_close(char *, buffer, bool);
int cmd_help(char *, bool);

char * paste=NULL; // pastebuffer, globally available

int main(int argc, char *argv[])
{
	char * loadfn = NULL;
	char * scriptfn = NULL;
	int arg;
	for (arg=1; arg<argc; arg++)
	{
		if((strcasecmp(argv[arg], "--version") == 0) || (strcasecmp(argv[arg], "-v") == 0))
		{ // print version info and exit
			printf(VERSION_MSG);
			return(0);
		}
		else if((strcasecmp(argv[arg], "--help") == 0) || (strcasecmp(argv[arg], "-?") == 0))
		{ // print invocation help and exit
			printf(USAGE_MSG);
			return(0);
		}
		else if((strncasecmp(argv[arg], "--script=", 9) == 0) || (strncasecmp(argv[arg], "-s=", 3) == 0))
		{ // script file
			char * eqsign=strchr(argv[arg], '=');
			if(eqsign==NULL)
			{
				printf("Internal error (\"impossible\"); -s option not handled\n");
			}
			else
			{
				scriptfn=eqsign+1;
			}
		}
		else
		{ // unrecognised option, assume it's a filename
			loadfn = argv[arg];
		}
	}
	
	printf(GPL_MSG);
	
	buffer * ebuf = (buffer *)malloc(sizeof(buffer));
	buffer sbuf; // script file
	sbuf.buf = NULL;
	sbuf.curline=0;
	int nbuffers=0, curbuffer=0;
	ebuf[0].buf = NULL;
	ebuf[0].nlines=0;
	ebuf[0].filename = NULL;
	ebuf[0].saved=true,
	ebuf[0].curline=0;
	bool cont=true;
	if(loadfn!=NULL)
	{
		printf("%s: ", loadfn);
		int f=open(loadfn, O_CREAT|O_RDWR, S_IFREG|S_IRUSR|S_IWUSR|S_IRGRP);
		close(f);
		ebuf[0]=cmd_load(loadfn, true);
		if(ebuf[0].err!=0)
		{
			return(ebuf[0].err);
		}
		nbuffers++;
	}
	bool scriptmode=false;
	if(scriptfn!=NULL)
	{
		printf("Script: ");
		sbuf=cmd_load(scriptfn, true);
		if(sbuf.err!=0)
		{
			return(sbuf.err);
		}
		scriptmode=true;
	}
	while(cont)
	{
		if(ebuf[curbuffer].curline>=ebuf[curbuffer].nlines)
			ebuf[curbuffer].curline=max(ebuf[curbuffer].nlines-1, 0);
		bool match=false;
		char * cmd;
		if(scriptmode)
		{
			if(sbuf.curline==sbuf.nlines)
			{
				cont=false;
				cmd=strdup("quit");
			}
			else
			{
				cmd=strdup(sbuf.buf[sbuf.curline++]);
			}
		}
		else
		{
			cmd = getl("(ted) ");
		}
		if(cmd==NULL)
		{
			printf("A strange error occurred\n");
			continue;
		}
		char * split = strchr(cmd, ' ');
		if(split==NULL)
		{
			split=cmd+strlen(cmd);
			split[1]=0;
		}
		*split++=0;
		while((strlen(cmd)!=0) && (cmd[strlen(cmd)-1]==' '))
			cmd[strlen(cmd)-1]=0;
		if(strcmp(cmd, "goto")==0)
		{
			if(scriptmode)
			{
				printf("Script: ");
				sbuf=cmd_move(split, sbuf);
				if(sbuf.err!=0)
				{
					printf("Error %u\n", sbuf.err);
					cont=false;
				}
			}
			else
			{
				printf("Not a script!\n");
			}
			match=true;
		}
		if(strcmp(cmd, "help")==0)
		{
			cmd_help(split, nbuffers==0);
			match=true;
		}
		if(strcmp(cmd, "new")==0)
		{
			buffer * newbuf=(buffer *)realloc(ebuf, (nbuffers+1)*sizeof(buffer));
			if(newbuf==NULL)
			{
				printf("A strange error occurred\n");
				if(scriptmode)
					cont=false;
			}
			else
			{
				ebuf=newbuf;
				ebuf[nbuffers]=cmd_new(split);
				curbuffer=nbuffers;
				if(scriptmode && ebuf[nbuffers].err!=0)
					cont=false;
				nbuffers++;
			}
			match=true;
		}
		if(strcmp(cmd, "load")==0)
		{
			buffer * newbuf=(buffer *)realloc(ebuf, (nbuffers+1)*sizeof(buffer));
			if(newbuf==NULL)
			{
				printf("A strange error occurred\n");
				if(scriptmode)
					cont=false;
			}
			else
			{
				ebuf=newbuf;
				ebuf[nbuffers]=cmd_load(split, true);
				curbuffer=nbuffers;
				if(scriptmode && ebuf[nbuffers].err!=0)
					cont=false;
				nbuffers++;
			}
			match=true;
		}
		if(strcmp(cmd, "quit")==0)
		{
			if(scriptmode)
				cont=false;
			else
			{
				if((nbuffers<2) && ((ebuf[0].buf==NULL) || ebuf[0].saved==true))
					cont=false;
				else
				{
					if(strchr(split, '!'))
						cont=false;
					else
					{
						printf("Unsaved changes, or multiple open buffers!  Use quit ! to quit anyway\n");
					}
				}
			}
			match=true;
		}
		if(nbuffers==0)
		{
			if(!match)
			{
				printf("No buffers open!\n");
				if(scriptmode)
					cont=false;
			}
		}
		else
		{
			if(strcmp(cmd, "buffer")==0)
			{
				/* 'buffer' command - switches between open buffers, or queries which buffer is open */
				if((strchr(split, '?')!=NULL) || (*split==0))
				{
					if(ebuf[curbuffer].filename!=NULL)
					{
						printf("Current: %sbuffer #%u of %u, '%s'\n", ebuf[curbuffer].saved?"":"*", curbuffer+1, nbuffers, ebuf[curbuffer].filename);
					}
					else
					{
						printf("Current: %sbuffer #%u of %u, untitled\n", ebuf[curbuffer].saved?"":"*", curbuffer+1, nbuffers);
					}
				}
				else
				{
					if(nbuffers<2)
					{
						printf("No other open buffers!\n");
						if(scriptmode)
							cont=false;
					}
					else
					{
						int newbuffer=-1;
						sscanf(split, "%u", &newbuffer);
						newbuffer--;
						if((newbuffer>=0) && (newbuffer<nbuffers))
						{
							curbuffer=newbuffer;
							if(ebuf[curbuffer].filename!=NULL)
							{
								printf("Selected buffer #%u '%s'\n", curbuffer+1, ebuf[curbuffer].filename);
							}
							else
							{
								printf("Selected buffer #%u (untitled)\n", curbuffer+1);
							}
							if(scriptmode && ebuf[curbuffer].err!=0)
								cont=false;
						}
						else
						{
							printf("Buffer number out of range!\n");
							if(scriptmode)
								cont=false;
						}
					}
				}
				match=true;
			}
			if(strcmp(cmd, "save")==0)
			{
				if(ebuf[curbuffer].buf==NULL)
				{
					printf("No file open!\n");
					if(scriptmode)
						cont=false;
				}
				else
				{
					ebuf[curbuffer]=cmd_save(split, ebuf[curbuffer]);
					if(scriptmode && ebuf[curbuffer].err!=0)
						cont=false;
				}
				match=true;
			}
			if(strcmp(cmd, "showall")==0)
			{
				if(ebuf[curbuffer].buf==NULL)
				{
					printf("No file open!\n");
					if(scriptmode)
						cont=false;
				}
				else
				{
					int e=cmd_show("-1,-1", ebuf[curbuffer]);
					if(scriptmode && e!=0)
						cont=false;
				}
				match=true;
			}
			if(strcmp(cmd, "showpg")==0)
			{
				if(ebuf[curbuffer].buf==NULL)
				{
					printf("No file open!\n");
					if(scriptmode)
						cont=false;
				}
				else
				{
					int e=cmd_showpg(ebuf[curbuffer]);
					if(scriptmode && e!=0)
						cont=false;
				}
				match=true;
			}
			if(strcmp(cmd, "show")==0)
			{
				if(ebuf[curbuffer].buf==NULL)
				{
					printf("No file open!\n");
					if(scriptmode)
						cont=false;
				}
				else
				{
					int e=cmd_show(split, ebuf[curbuffer]);
					if(scriptmode && e!=0)
						cont=false;
				}
				match=true;
			}
			if(strcmp(cmd, "move")==0)
			{
				if(ebuf[curbuffer].buf==NULL)
				{
					printf("No file open!\n");
					if(scriptmode)
						cont=false;
				}
				else
				{
					ebuf[curbuffer]=cmd_move(split, ebuf[curbuffer]);
					if(scriptmode && ebuf[curbuffer].err!=0)
						cont=false;
				}
				match=true;
			}
			if(strcmp(cmd, "insert")==0)
			{
				if(ebuf[curbuffer].buf==NULL)
				{
					printf("No file open!\n");
					if(scriptmode)
						cont=false;
				}
				else
				{
					ebuf[curbuffer]=cmd_insert(split, ebuf[curbuffer]);
					if(scriptmode && ebuf[curbuffer].err!=0)
						cont=false;
				}
				match=true;
			}
			if(strcmp(cmd, "copy")==0)
			{
				if(ebuf[curbuffer].buf==NULL)
				{
					printf("No file open!\n");
					if(scriptmode)
						cont=false;
				}
				else
				{
					int e=cmd_copy(split, ebuf[curbuffer]);
					if(scriptmode && e!=0)
						cont=false;
				}
				match=true;
			}
			if(strcmp(cmd, "paste")==0)
			{
				if(ebuf[curbuffer].buf==NULL)
				{
					printf("No file open!\n");
					if(scriptmode)
						cont=false;
				}
				else
				{
					ebuf[curbuffer]=cmd_paste(split, ebuf[curbuffer]);
					if(scriptmode && ebuf[curbuffer].err!=0)
						cont=false;
				}
				match=true;
			}
			if(strcmp(cmd, "replace")==0)
			{
				if(ebuf[curbuffer].buf==NULL)
				{
					printf("No file open!\n");
					if(scriptmode)
						cont=false;
				}
				else
				{
					ebuf[curbuffer]=cmd_replace(split, ebuf[curbuffer]);
					if(scriptmode && ebuf[curbuffer].err!=0)
						cont=false;
				}
				match=true;
			}
			if(strcmp(cmd, "split")==0)
			{
				if(ebuf[curbuffer].buf==NULL)
				{
					printf("No file open!\n");
					if(scriptmode)
						cont=false;
				}
				else
				{
					ebuf[curbuffer]=cmd_split(split, ebuf[curbuffer]);
					if(scriptmode && ebuf[curbuffer].err!=0)
						cont=false;
				}
				match=true;
			}
			if(strcmp(cmd, "append")==0)
			{
				if(ebuf[curbuffer].buf==NULL)
				{
					printf("No file open!\n");
					if(scriptmode)
						cont=false;
				}
				else
				{
					ebuf[curbuffer]=cmd_append(split, ebuf[curbuffer]);
					if(scriptmode && ebuf[curbuffer].err!=0)
						cont=false;
				}
				match=true;
			}
			if(strcmp(cmd, "delete")==0)
			{
				if(ebuf[curbuffer].buf==NULL)
				{
					printf("No file open!\n");
					if(scriptmode)
						cont=false;
				}
				else
				{
					ebuf[curbuffer]=cmd_delete(split, ebuf[curbuffer]);
					if(scriptmode && ebuf[curbuffer].err!=0)
						cont=false;
				}
				match=true;
			}
			if(strcmp(cmd, "join")==0)
			{
				if(ebuf[curbuffer].buf==NULL)
				{
					printf("No file open!\n");
					if(scriptmode)
						cont=false;
				}
				else
				{
					ebuf[curbuffer]=cmd_join(split, ebuf[curbuffer]);
					if(scriptmode && ebuf[curbuffer].err!=0)
						cont=false;
				}
				match=true;
			}
			if(strcmp(cmd, "close")==0)
			{
				if(ebuf[curbuffer].buf==NULL)
				{
					printf("No file open!\n");
					if(scriptmode)
						cont=false;
				}
				else
				{
					ebuf[curbuffer]=cmd_close(split, ebuf[curbuffer], true);
					if(ebuf[curbuffer].buf==NULL)
					{
						nbuffers--;
						int i;
						for(i=curbuffer;i<nbuffers;i++)
						{
							ebuf[i]=ebuf[i+1];
						}
						if(curbuffer>=nbuffers)
							curbuffer=max(nbuffers-1, 0);
						if(nbuffers!=0)
						{
							buffer * newbuf=(buffer *)realloc(ebuf, nbuffers*sizeof(buffer));
							if(newbuf==NULL)
							{
								printf("A strange (non-critical) error occurred\n");
							}
							else
							{
								ebuf=newbuf;
							}
						}
					}
				}
				match=true;
			}
			if(strcmp(cmd, "count")==0)
			{
				if(ebuf[curbuffer].buf==NULL)
				{
					printf("No file open!\n");
					if(scriptmode)
						cont=false;
				}
				else
				{
					int e=cmd_count(split, ebuf[curbuffer]);
					if(scriptmode && e!=0)
						cont=false;
				}
				match=true;
			}
			if(!match)
			{
				printf("Command %s not recognised\n", cmd);
				if(scriptmode)
					cont=false;
			}
		}
		free(cmd);
	}
	
	return(0);
}

char * getl(char * prompt)
{
	printf("%s", prompt);
	fflush(stdout);
	// gets a line of string data, {re}alloc()ing as it goes, so you don't need to make a buffer for it, nor must thee fret thyself about overruns!
	char * lout = (char *)malloc(80);
	int i=0;
	char c;
	while(1)
	{
		c = getchar();
		if (c == 10)
			break;
		if (c != 0)
			lout[i++]=c;
		if ((i%80) == 0)
		{
			if ((lout = (char *)realloc(lout, i+80))==NULL)
			{
				free(lout);
				printf("\nNot enough memory to store input!\n");
				return(NULL);
			}
		}
	}
	lout[i]=0;
	lout=(char *)realloc(lout, i+1);
	return(lout);
}

char * slurp(FILE *fp)
{
	fseek(fp, 0, SEEK_SET);
	// slurps a filefull of textual data, {re}alloc()ing as it goes, so you don't need to make a buffer for it, nor must thee fret thyself about overruns!
	char * lout = (char *)malloc(80);
	int i=0;
	int c;
	while(!feof(fp))
	{
		c = fgetc(fp);
		if(c == EOF)
			break;
		if (c != 0)
			lout[i++]=(char)c;
		if ((i%80) == 0)
		{
			if ((lout = (char *)realloc(lout, i+80))==NULL)
			{
				free(lout);
				fprintf(stderr, "\nNot enough memory to store input!\n");
				return(NULL);
			}
		}
	}
	lout[i]=0;
	lout=(char *)realloc(lout, i+1);
	return(lout);
}

buffer cmd_new(char * split)
{
	/* 'new' command, creates a new buffer */
	buffer rv;
	rv.nlines=0;
	rv.buf=(char **)malloc(rv.nlines*sizeof(char *));
	if(rv.buf==NULL)
	{
		printf("A strange error occurred\n");
		rv.err=1;
		return(rv);
	}
	rv.filename=strdup(split);
	rv.saved=true;
	rv.curline=0;
	rv.err=0;
	return(rv);
}

buffer cmd_load(char * split, bool echo)
{
	/* 'load' command, loads a file from disk into a new buffer */
	buffer ebuf;
	FILE *fp=fopen(split, "r");
	if(fp==NULL)
	{
		if(echo)
			printf("Failed to open file!\n");
		ebuf.err=1;
	}
	else
	{
		char * data = slurp(fp);
		fclose(fp);
		if(data==NULL)
		{
			if(echo)
				printf("Failed to load file!\n");
			ebuf.err=2;
		}
		else
		{
			ebuf.nlines=0;
			char * o=data;
			while(o!=NULL)
			{
				o=strchr(o+1, '\n');
				ebuf.nlines++;
			}
			ebuf.buf=(char **)malloc(ebuf.nlines*sizeof(char *));
			int line=0;
			o=data-1;
			while(o!=NULL)
			{
				char * next=strchr(o+1, '\n');
				if(next==NULL)
				{
					ebuf.buf[line]=strdup(o);
				}
				else
				{
					ebuf.buf[line]=(char *)malloc((size_t)(next-o));
					strncpy(ebuf.buf[line], o+1, next-o-1);
					ebuf.buf[line][next-o-1]=0;
					line++;
				}
				o=next;
			}
			ebuf.nlines--;
			ebuf.filename=strdup(split);
			if(echo)
				printf("Read %u bytes, %u lines\n", strlen(data), ebuf.nlines);
			free(data);
			ebuf.curline=0;
			ebuf.err=0;
			ebuf.saved=true;
		}
	}
	return(ebuf);
}

buffer cmd_save(char * split, buffer ebuf)
{
	/* 'save' command, writes a buffer out to disk file */
	if(strlen(split)>0)
		sscanf(split, "%s", ebuf.filename);
	if((ebuf.filename==NULL) || (*ebuf.filename==0))
	{
		printf("No filename specified!\n");
		ebuf.err=2;
		return(ebuf);
	}
	FILE *fp = fopen(ebuf.filename, "w");
	if(fp==NULL)
	{
		printf("Failed to open file '%s' for writing\n", ebuf.filename);
		ebuf.err=1;
	}
	else
	{
		int line;
		size_t bytes=0;
		for(line=0;line<ebuf.nlines;line++)
		{
			int p=0;
			while(ebuf.buf[line][p]!=0)
			{
				fputc(ebuf.buf[line][p], fp);
				bytes++;
				p++;
			}
			fputc('\n', fp);
		}
		printf("Wrote %zu bytes, %u lines\n", bytes, ebuf.nlines);
		fclose(fp);
		ebuf.saved=true;
		ebuf.err=0;
	}
	return(ebuf);
}

int cmd_show(char * split, buffer ebuf)
{
	/* 'show' command, displays current line (and context) of buffer on-screen */
	int showline=-1;
	int context=2;
	sscanf(split, "%d,%d", &showline, &context); // if there isn't one, it should leave it unchanged
	if(showline==-1)
		showline=ebuf.curline;
	if(context==-1)
		context=ebuf.nlines;
	printf("LINE:|   |   | 0 |   |   0   |   | 0 |   |   0   |   | 0 |   |   0   |   | 0 |\n");
	int dline;
	for(dline=-context;dline<context+1;dline++)
	{
		int line=showline+dline;
		if((line>=0) && (line<ebuf.nlines))
		{
			printf("%04u%c", line, line==ebuf.curline?'#':':');
			int p;
			for(p=0;ebuf.buf[line][p]!=0;p++)
			{
				if(ebuf.buf[line][p]=='\t')
					printf("    ");
				else
					printf("%c", ebuf.buf[line][p]);
			}
			printf("$\n");
		}
	}
	if(context==ebuf.nlines)
		printf("LINE:|   |   | 0 |   |   0   |   | 0 |   |   0   |   | 0 |   |   0   |   | 0 |\n");
	return(0);
}

int cmd_showpg(buffer ebuf)
{
	/* 'showpg' command, shows the whole buffer, one page at a time */
	int line;
	int slines=0;
	int scrlines=24;
	int scrcols=80;
	char * envlines = getenv("LINES"), * envcols = getenv("COLUMNS");
	if(envlines!=NULL)
		sscanf(envlines, "%u", &scrlines);
	if(envcols!=NULL)
		sscanf(envcols, "%u", &scrcols);
	if(scrlines<=0)
		scrlines=24;
	if(scrcols<=0)
		scrcols=80;
	printf("LINE:|   |   | 0 |   |   0   |   | 0 |   |   0   |   | 0 |   |   0   |   | 0 |\n");
	bool cont=true;
	for(line=0;line<ebuf.nlines;line++)
	{
		printf("%04u%c", line, line==ebuf.curline?'#':':');
		int schars=5;
		int p;
		for(p=0;ebuf.buf[line][p]!=0;p++)
		{
			if(ebuf.buf[line][p]=='\t')
			{
				printf("    ");
				schars+=4;
			}
			else
			{
				printf("%c", ebuf.buf[line][p]);
				schars++;
			}
			if(schars>=scrcols)
			{
				schars-=scrcols;
				slines++;
			}
			if(slines+3>scrlines)
			{
				printf("\nLINE:|   |   | 0 |   |   0   |   | 0 |   |   0   |   | 0 |   |   0   |   | 0 |");
				fflush(stdout);
				char c;
				while(1)
				{
					c=getchar();
					if(c=='\n')
						break;
					if(c=='x')
						cont=false;
				}
				slines=0;
			}
			if(!cont)
				break;
		}
		if(cont)
			printf("$\n");
		slines++;
		if(slines+3>scrlines)
		{
			printf("LINE:|   |   | 0 |   |   0   |   | 0 |   |   0   |   | 0 |   |   0   |   | 0 |");
			fflush(stdout);
			char c;
			while(1)
			{
				c=getchar();
				if(c=='\n')
					break;
				if(c=='x')
					cont=false;
			}
			slines=0;
		}
		if(!cont)
			break;
	}
	return(0);
}

buffer cmd_move(char * split, buffer ebuf)
{
	/* 'move' command, sets the current line */
	char movetype='=';
	int moveline=0;
	if((*split=='=') || (*split=='+') || (*split=='-') || (*split=='$'))
	{
		movetype=*split++;
	}
	sscanf(split, "%u", &moveline);
	switch(movetype)
	{
		case '=':
		break;
		case '+':
			moveline+=ebuf.curline;
		break;
		case '-':
			moveline=ebuf.curline-moveline;
		break;
		case '$':
			moveline=ebuf.nlines-1;
	}
	if(moveline<ebuf.nlines)
	{
		ebuf.curline=moveline;
		printf("Moved to line %u\n", ebuf.curline);
		ebuf.err=0;
	}
	else
	{
		printf("Line number out of file\n");
		ebuf.err=1;
	}
	return(ebuf);
}

buffer cmd_insert(char * split, buffer ebuf)
{
	/* 'insert' command, adds one or more new lines to the buffer */
	bool rpt=(strchr(split, '.')!=NULL);
	bool first=true;
	if(rpt)
		printf("To end, enter a dot (.) on a line by itself\n");
	char * line;
	insert:
	line=getl("(ins) ");
	if(line==NULL)
	{
		printf("A strange error occurred\n");
		ebuf.err=1;
	}
	else
	{
		if(rpt && (strcmp(line, ".")==0))
		{
			printf("Line(s) inserted\n");
			free(line);
			ebuf.err=0;
		}
		else
		{
			char ** newbuf=(char **)realloc(ebuf.buf, (1+ebuf.nlines)*sizeof(char *));
			if(newbuf==NULL)
			{
				printf("A strange error occurred\n");
				ebuf.err=1;
			}
			else
			{
				if(strchr(split,'+') && first && ebuf.curline<ebuf.nlines)
					ebuf.curline++;
				first=false;
				ebuf.buf=newbuf;
				int i;
				for(i=ebuf.nlines;i>ebuf.curline;i--)
				{
					ebuf.buf[i]=ebuf.buf[i-1];
				}
				ebuf.buf[ebuf.curline]=strdup(line);
				free(line);
				ebuf.nlines++;
				ebuf.saved=false;
				if(rpt)
				{
					ebuf.curline++;
					goto insert;
				}
				printf("Line inserted\n");
				ebuf.err=0;
			}
		}
	}
	return(ebuf);
}

buffer cmd_replace(char * split, buffer ebuf)
{
	/* 'replace' command, overwrites the current line (but stores the old line in the pastebuffer) */
	if(ebuf.nlines==0)
	{
		printf("The current file is empty!\n");
		return(ebuf);
	}
	char * line=getl("(rep) ");
	if(line==NULL)
	{
		printf("A strange error occurred\n");
		ebuf.err=1;
	}
	else
	{
		if(paste)
			free(paste);
		paste=ebuf.buf[ebuf.curline];
		ebuf.buf[ebuf.curline]=line;
		ebuf.saved=false;
		printf("Line replaced\n");
		ebuf.err=0;
	}
	return(ebuf);
}

buffer cmd_append(char * split, buffer ebuf)
{
	/* 'append' command, adds a line to the end of the buffer */
	char * line=getl("(app) ");
	if(line==NULL)
	{
		printf("A strange error occurred\n");
		ebuf.err=1;
	}
	else
	{
		char ** newbuf=(char **)realloc(ebuf.buf, (1+ebuf.nlines)*sizeof(char *));
		if(newbuf==NULL)
		{
			printf("A strange error occurred\n");
			ebuf.err=1;
		}
		else
		{
			ebuf.buf=newbuf;
			ebuf.buf[ebuf.nlines]=line;
			ebuf.nlines++;
			ebuf.saved=false;
			printf("Line appended\n");
			ebuf.err=0;
		}
	}
	return(ebuf);
}

buffer cmd_split(char * split, buffer ebuf)
{
	/* 'split' command, splits the current line into two */
	if(ebuf.nlines==0)
	{
		printf("The current file is empty!\n");
		return(ebuf);
	}
	int pos=0;
	sscanf(split, "%d", &pos);
	int p=0,q=0;
	if(pos>=0)
	{
		for(q=0;q<=strlen(ebuf.buf[ebuf.curline]);q++)
		{
			if(ebuf.buf[ebuf.curline][q]=='\t')
				p+=4;
			else
				p++;
			if(p>pos)
				break;
		}
	}
	else
	{
		p++;
		for(q=strlen(ebuf.buf[ebuf.curline])-1;q>=0;q--)
		{
			if(ebuf.buf[ebuf.curline][q]=='\t')
				p+=4;
			else
				p++;
			if(p>-pos)
				break;
		}
	}
	if(q>=strlen(ebuf.buf[ebuf.curline]))
	{
		printf("Column number out of line\n");
		ebuf.err=1;
	}
	else
	{
		char * line=strdup(ebuf.buf[ebuf.curline]+q);
		if(line==NULL)
		{
			printf("A strange error occurred\n");
			ebuf.err=2;
		}
		else
		{
			char ** newbuf=(char **)realloc(ebuf.buf, (1+ebuf.nlines)*sizeof(char *));
			if(newbuf==NULL)
			{
				printf("A strange error occurred\n");
				ebuf.err=2;
			}
			else
			{
				ebuf.buf[ebuf.curline][q]=0;
				ebuf.buf[ebuf.curline]=(char *)realloc(ebuf.buf[ebuf.curline], max(q, 1));
				ebuf.curline++;
				ebuf.buf=newbuf;
				int i;
				for(i=ebuf.nlines;i>ebuf.curline;i--)
				{
					ebuf.buf[i]=ebuf.buf[i-1];
				}
				ebuf.buf[ebuf.curline]=line;
				ebuf.nlines++;
				ebuf.saved=false;
				printf("Line split\n");
				ebuf.err=0;
			}
		}
	}
	return(ebuf);
}

buffer cmd_join(char * split, buffer ebuf)
{
	/* 'join' command, joins two lines together */
	if(ebuf.nlines<2)
	{
		printf("Not enough lines to join!\n");
		return(ebuf);
	}
	if((strchr(split, '+')==NULL) || ebuf.curline+1==ebuf.nlines)
	{
		if(ebuf.curline>0)
			ebuf.curline--;
	}
	ebuf.buf[ebuf.curline]=(char *)realloc(ebuf.buf[ebuf.curline], strlen(ebuf.buf[ebuf.curline])+strlen(ebuf.buf[ebuf.curline+1])+1);
	strcat(ebuf.buf[ebuf.curline], ebuf.buf[ebuf.curline+1]);
	free(ebuf.buf[ebuf.curline+1]);
	int i;
	ebuf.nlines--;
	for(i=ebuf.curline+1;i<ebuf.nlines;i++)
	{
		ebuf.buf[i]=ebuf.buf[i+1];
	}
	ebuf.saved=false;
	if(ebuf.nlines>0)
	{
		char ** newbuf=(char **)realloc(ebuf.buf, (ebuf.nlines)*sizeof(char *));
		if(newbuf==NULL)
			printf("A strange error occurred\n");
		else
			ebuf.buf=newbuf;
	}
	printf("Lines joined\n");
	ebuf.err=0;
	return(ebuf);
}

buffer cmd_delete(char * split, buffer ebuf)
{
	/* 'delete' command, removes a line from the buffer (but stores it in the pastebuffer) */
	if(ebuf.nlines==0)
	{
		printf("The current file is empty!\n");
		return(ebuf);
	}
	if(paste)
		free(paste);
	paste=ebuf.buf[ebuf.curline];
	int i;
	ebuf.nlines--;
	for(i=ebuf.curline;i<ebuf.nlines;i++)
	{
		ebuf.buf[i]=ebuf.buf[i+1];
	}
	ebuf.saved=false;
	if(ebuf.nlines>0)
	{
		char ** newbuf=(char **)realloc(ebuf.buf, (ebuf.nlines)*sizeof(char *));
		if(newbuf==NULL)
			printf("A strange (but non-critical) error occurred\n");
		else
			ebuf.buf=newbuf;
	}
	printf("Line deleted\n");
	ebuf.err=0;
	return(ebuf);
}

int cmd_copy(char * split, buffer ebuf)
{
	/* 'copy' command, copies the current line into the pastebuffer */
	if(ebuf.nlines==0)
	{
		printf("The current file is empty!\n");
		return(1);
	}
	if(paste)
		free(paste);
	paste=strdup(ebuf.buf[ebuf.curline]);
	return(0);
}

buffer cmd_paste(char * split, buffer ebuf)
{
	/* 'paste' command, inserts the contents of the pastebuffer at the current line */
	if(paste==NULL)
	{
		printf("There is nothing to paste\n");
		ebuf.err=2;
	}
	else
	{
		char ** newbuf=(char **)realloc(ebuf.buf, (1+ebuf.nlines)*sizeof(char *));
		if(newbuf==NULL)
		{
			printf("A strange error occurred\n");
			ebuf.err=1;
		}
		else
		{
			if(strchr(split,'+') && ebuf.nlines!=0)
				ebuf.curline++;
			ebuf.buf=newbuf;
			int i;
			for(i=ebuf.nlines;i>ebuf.curline;i--)
			{
				ebuf.buf[i]=ebuf.buf[i-1];
			}
			ebuf.buf[ebuf.curline]=strdup(paste);
			ebuf.nlines++;
			ebuf.saved=false;
			printf("Line pasted\n");
			ebuf.err=0;
		}
	}
	return(ebuf);
}

int cmd_count(char * split, buffer ebuf)
{
	/* 'count' command, counts bytes, words and lines */
	int words=0;
	size_t bytes=0;
	int line;
	for(line=0;line<ebuf.nlines;line++)
	{
		bytes+=strlen(ebuf.buf[line])+1;
		char * o=strdup(ebuf.buf[line]);
		char * p=strtok(o, " \t");
		while(p!=NULL)
		{
			words++;
			p=strtok(NULL, " \t");
		}
		free(o);
	}
	printf("Count: %zu bytes, %u words, %u lines\n", bytes, words, ebuf.nlines);
	return(0);
}

buffer cmd_close(char * split, buffer ebuf, bool echo)
{
	/* 'close' command, closes a buffer */
	if(strchr(split, '!') || ebuf.saved)
	{
		if(ebuf.buf)
		{
			int i;
			for(i=0;i<ebuf.nlines;i++)
				free(ebuf.buf[i]);
			free(ebuf.buf);
		}
		ebuf.buf=NULL;
		if(ebuf.filename)
			free(ebuf.filename);
		ebuf.filename=NULL;
		if(echo)
			printf("Closed file\n");
		ebuf.err=0;
	}
	else
	{
		if(echo)
			printf("Unsaved changes!  Use close ! to close anyway\n");
		ebuf.err=1;
	}
	return(ebuf);
}

int cmd_help(char * split, bool nobufs)
{
	/* 'help' command, provides on-line documentation */
	int rv=0;
	if(*split!=0)
	{
		buffer hbuf=cmd_load(PREFIX"/share/doc/ted/cmdhelp", false);
		if(hbuf.err)
		{
			printf("It appears that you do not have ted command help installed.\n");
			rv=128+hbuf.err;
		}
		else
		{
			int i,j=0;
			for(i=0;i<hbuf.nlines;i++)
			{
				char * line=hbuf.buf[i];
				char * colon=strchr(line, ':');
				*colon++=0;
				char * ptr=strchr(colon, '~');
				while(ptr!=NULL)
				{
					*ptr++='\n';
					ptr=strchr(ptr, '~');
				}
				if(strcmp(split, line)==0)
				{
					printf("%s: %s\n", line, colon);
					j++;
				}
			}
			if(j==0)
			{
				printf("[help]: no such command '%s'\n", split);
				rv=128;
			}
		}
		cmd_close("", hbuf, false);
	}
	else
	{
		buffer hbuf=cmd_load(PREFIX"/share/doc/ted/readme", false);
		if(hbuf.err)
		{
			printf("It appears that you do not have ted help installed.\n");
			rv=128+hbuf.err;
		}
		else
		{
			cmd_showpg(hbuf);
			cmd_close("", hbuf, false);
		}
	}
	return(rv);
}
